"""Robot Framework.

The Driver module contains a Driver class which executes a run and a
loaddriver() factory method for creating driver instances based on the Robot
configuration.

The Action module contains an IAction interface that any action to be executed
by Driver should implement.

The Utility module contains utility functions used throughout the robot code.

"""